""" Configuration module for the broker. """
from typing import Optional

from .utils import BaseModel


class Config(BaseModel):
    """
    Configuration class for SSI datafeed information.
    Attributes:
        ssi_broker_account (Optional[str]): The SSI broker account.
        ssi_broker_id (Optional[str]): The SSI broker ID.
        ssi_broker_secret (Optional[str]): The SSI broker secret.
        ssi_broker_private_key (Optional[str]): The SSI broker private key.
    """
    # SSI datafeed information
    ssi_broker_account: Optional[str] = None
    ssi_broker_id: Optional[str] = None
    ssi_broker_secret: Optional[str] = None
    ssi_broker_private_key: Optional[str] = None



from datetime import datetime, timedelta


def split_date_range(start_date, end_date, days_interval=7):
    """
    Splits a date range into smaller intervals of a specified number of days.
    Args:
        start_date (str): The start date of the range in 'YYYY-MM-DD' format.
        end_date (str): The end date of the range in 'YYYY-MM-DD' format.
        days_interval (int, optional): The number of days for each interval. Defaults to 7.
    Returns:
        list of tuple: A list of tuples where each tuple contains the start and end date of an
        interval in 'YYYY-MM-DD' format.
    """
    current = datetime.strptime(start_date, '%Y-%m-%d')
    end = datetime.strptime(end_date, '%Y-%m-%d')

    intervals = []

    while current < end:
        interval_end = min(current + timedelta(days=days_interval-1), end)
        intervals.append((
            current.strftime('%Y-%m-%d'),
            interval_end.strftime('%Y-%m-%d')
        ))
        current = interval_end + timedelta(days=1)
    return intervals


def convert_timestamp_to_datetime(timestamp: int):
    """
    Converts a timestamp to a date in 'YYYY-MM-DD' format.
    Args:
        timestamp (int): The timestamp to convert.
    Returns:
        str: The date in 'YYYY-MM-DD' format.
    """
    return datetime.fromtimestamp(timestamp).strftime('%Y-%m-%d %H:%M:%S')



""" Interface for Broker API """
from typing import List
from abc import ABC, abstractmethod
from .config import Config
from .model import vBrokerOrder


class IBrokerAPI(ABC):
    def __init__(self, config: Config):
        self.config: Config = config
        self.otp: str = None

    @property
    def otp(self) -> str:
        return self._otp

    @otp.setter
    def otp(self, value: str):
        self._otp = value

    @abstractmethod
    def get_otp(self) -> str:
        return NotImplemented

    @abstractmethod
    def get_token(self) -> str:
        return NotImplemented

    @abstractmethod
    def get_max_buy_quantity(self, account_no: str, instrument: str, price: float) -> dict:
        return NotImplemented

    @abstractmethod
    def get_max_sell_quantity(self, account_no: str, instrument: str, price: float) -> dict:
        return NotImplemented

    @abstractmethod
    def get_ordebbook(self, account_no: str, from_date: str, to_date: str) -> List[vBrokerOrder]:
        return NotImplemented

    @abstractmethod
    def get_positions(self, account_no: str, is_equity: bool = True) -> dict:
        return NotImplemented

    @abstractmethod
    def get_balance(self, account_no: str, is_equity: bool = True) -> dict:
        return NotImplemented

    @abstractmethod
    def place_order(
        self, account_no: str, side: str, instrument: str,
        quantity: int, price: float | str, is_equity: bool = True
    ) -> dict:
        return NotImplemented

    @abstractmethod
    def modify_order(
        self, account_no: str, order_id: str,
        side: str, instrument: str, quantity: int, price: float | str, is_equity: bool = True
    ) -> dict:
        return NotImplemented

    @abstractmethod
    def cancel_order(
        self, account_no: str, order_id: str, instrument: str, side: str, is_equity: bool = True
    ) -> dict:
        return NotImplemented



from enum import Enum


class EnumHandler(Enum):
    def __str__(self) -> str:
        return str.__str__(self)

    @classmethod
    def keys(cls):
        return cls._member_names_

    @classmethod
    def values(cls):
        return list(cls._value2member_map_.keys())

    @classmethod
    def items(cls):
        return list(cls)



"""
This module contains the EnumDatafeed class which is used for handling datafeed enumerations.
"""
from ..utils import EnumHandler


class SSIOrderStatusEnum(EnumHandler):
    """Order status enum"""

    WA = "WAITING"
    RS = "WAITING"
    SD = "WAITING"
    QU = "WAITING"
    IAV = "WAITING"
    WM = "WAITING"
    WC = "WAITING"
    CL = "CANCELED"
    PF = "PARTIALLY_FILLED"
    FFPC = "PARTIALLY_FILLED"
    FF = "FILLED"
    EX = "EXPIRED"
    RJ = "REJECTED"


class SSISideEnum(EnumHandler):
    """Side enum"""

    B = "BUY"
    S = "SELL"
    buy = "B"
    sell = "S"



""" Broker module for broker API and HUB. """
from .config import Config

from .enum_broker import EnumBroker
from .interface_broker_api import IBrokerAPI
from .interface_broker_hub import IBrokerHUB

from .ssi import SSIBrokerAPI, SSIBrokerHUB


class Broker:
    def __init__(self, broker: EnumBroker, config: Config) -> None:
        if broker == EnumBroker.SSI.value:
            self.__api: IBrokerAPI = SSIBrokerAPI(config)
            self.__hub: IBrokerHUB = SSIBrokerHUB(self.__api)

    @property
    def api(self) -> IBrokerAPI:
        return self.__api

    @property
    def hub(self) -> IBrokerHUB:
        return self.__hub



""" Broker module. """
from .enum_broker import EnumBroker  # noqa: F401
from .broker import Broker  # noqa: F401
from .config import Config  # noqa: F401
from .model import vBrokerOrder  # noqa: F401



import jwt
import time


class JWTHandler:
    def is_expired(self, bearer_token: str) -> bool:
        if bearer_token is None:
            return True
        decoded = jwt.decode(bearer_token.replace("Bearer ", ""), options={"verify_signature": False})
        return int(time.time()) > (decoded.get("exp") - 1)


jwt_handler = JWTHandler()



import json
import asyncio
import random
from urllib.parse import urlencode

from .constant import HUB_URL, HUB
from .enum import SSIOrderStatusEnum, SSISideEnum
from ..interface_broker_hub import IBrokerHUB
from ..utils import SocketListener, request_handler, convert_timestamp_to_datetime
from ..model import vBrokerOrder
from ..enum_broker import OrderStatusEnum, SideEnum


class SSIBrokerHUB(IBrokerHUB):
    def __init__(self, api):
        super().__init__(api)
        self.url: str = HUB_URL.replace("wss", "https")
        self.url_hub: str = HUB_URL
        self.message_send_to_socket: dict = {
            "H": HUB,
            "M": "SwitchChannels",
            "I": 0,
        }
        # Reconnection settings
        self.max_reconnect_attempts = 5
        self.base_delay = 1  # Base delay in seconds
        self.max_delay = 60  # Maximum delay between reconnection attempts

    def generate_socket_url(self):
        """
        Generates the socket URL for the connection.
        Returns:
            str: The socket URL.
        """
        self.connection_data: dict = {
            "connectionData": '[{"name": "' + HUB + '"}]',
            "clientProtocol": '1.3'
        }
        self.negotiate_query = urlencode(self.connection_data)
        self.url_negotiate = f"{self.url}/negotiate?{self.negotiate_query}"
        response = request_handler.post(self.url_negotiate, headers=self.headers)
        query = urlencode({
            "transport": "webSockets",
            "connectionToken": response["ConnectionToken"],
            "connectionData": '[{"name": "' + HUB + '"}]',
            "clientProtocol": response["ProtocolVersion"],
        })
        socket_url = f"{self.url_hub}/connect?{query}"
        return socket_url

    def calculate_backoff_delay(self, attempt):
        """
        Calculate exponential backoff delay with jitter.
        Args:
            attempt (int): Current reconnection attempt number.
        Returns:
            float: Delay in seconds before next reconnection attempt.
        """
        # Exponential backoff with full jitter
        delay = min(self.max_delay, self.base_delay * (2 ** attempt))
        jitter = random.uniform(0, delay)
        return jitter

    async def listen(self, on_message):
        """
        Listens for messages from the socket server with automatic reconnection.
        Args:
            on_message: The callback function to handle incoming messages.
        """
        self.headers: dict = {
            "Authorization": self.api.get_token(),
        }
        for attempt in range(self.max_reconnect_attempts):
            try:
                socket = SocketListener()
                async with socket.connect_socket_server(
                    self.generate_socket_url(), self.headers
                ) as websocket:
                    print(f"[vBroker] WebSocket connected (Attempt {attempt + 1})")
                    self.max_reconnect_attempts = 5
                    async for msg in websocket:
                        try:
                            msg = json.loads(msg)
                            if "M" not in msg:
                                continue
                            for i in msg["M"]:
                                if "A" not in i or not len(i["A"]):
                                    continue
                                msg = json.loads(i["A"][0])
                                if not msg.get("data"):
                                    continue
                                message = msg.get("data").get("message")
                                if msg.get("data").get("orderStatus"):
                                    _raw_status = msg.get("data").get("orderStatus")
                                    status = OrderStatusEnum[SSIOrderStatusEnum[_raw_status].value]
                                    if not message:
                                        message = _raw_status
                                    cancelled_quantity = msg.get("data").get("cancelQty", 0)
                                else:
                                    status = OrderStatusEnum.REJECTED
                                    cancelled_quantity = msg.get("data").get("quantity")
                                on_message(vBrokerOrder(
                                    account_no=msg.get("data").get("account"),
                                    order_id=msg.get("data").get("orderID"),
                                    unique_id=msg.get("data").get("origRequestID"),
                                    instrument=msg.get("data").get("instrumentID"),
                                    side=SideEnum[
                                        SSISideEnum[msg.get("data").get("buySell")].value
                                    ],
                                    order_type='',
                                    price=msg.get("data").get("price", 0),
                                    avg_price=msg.get("data").get("avgPrice", 0),
                                    quantity=msg.get("data").get("quantity"),
                                    filled_quantity=msg.get("data").get("filledQty", 0),
                                    os_quantity=msg.get("data").get("osQty", 0),
                                    cancelled_quantity=cancelled_quantity,
                                    status=status,
                                    input_time=convert_timestamp_to_datetime(
                                        int(msg.get("data").get("inputTime"))/1000
                                    ),
                                    modified_time=convert_timestamp_to_datetime(
                                        int(msg.get("data").get("modifiedTime"))/1000
                                    ),
                                    message=message
                                ))
                        except Exception as e:
                            print(f"[vBroker] Message processing error: {e}")

            except Exception as e:
                print(f"[vBroker] Connection error: {e}")
                # If this was the last attempt, raise the exception
                if attempt == self.max_reconnect_attempts - 1:
                    raise
                # Calculate backoff delay
                delay = self.calculate_backoff_delay(attempt)
                print(f"[vBroker] Reconnecting in {delay:.2f} seconds...")
                # Wait before trying to reconnect
                await asyncio.sleep(delay)
        print("[vBroker] Maximum reconnection attempts reached")



""" This module is responsible for handling the socket connection to the server. """
import websockets


class SocketListener:
    """
    A class that handles socket connections.
    Methods:
    --------
    connect_socket_server(url: str, headers: dict) -> websockets.WebSocketClientProtocol:
        Connects to a socket server using the specified URL and headers.
    """
    def connect_socket_server(self, url: str, headers: dict) -> websockets.WebSocketClientProtocol:
        """
        Connects to a socket server using the specified URL and headers.
        Parameters:
        -----------
        url : str
            The URL of the socket server to connect to.
        headers : dict
            Additional headers to include in the connection request.
        Returns:
        --------
        websockets.WebSocketClientProtocol
            The WebSocket client protocol object representing the connection.
        """
        return websockets.connect(url, extra_headers=headers)



"""
This module contains the EnumDatafeed class which is used for handling datafeed enumerations.
"""
from .utils import EnumHandler


class EnumBroker(EnumHandler):
    """
    This module contains the EnumDatafeed class which is used for handling datafeed enumerations.
    Attributes:
        SSI (str): Represents the SSI datafeed.
    """
    SSI = 'ssi'
    SSI_PAPER = 'ssi_paper'


class OrderStatusEnum(EnumHandler):
    """Order status enum"""

    PREPARE = "prepare"
    WAITING = "waiting"
    PARTIALLY_FILLED = "partially_filled"
    FILLED = "filled"
    CANCELED = "canceled"
    REJECTED = "rejected"
    EXPIRED = "expired"


class SideEnum(EnumHandler):
    """Side enum"""

    BUY = "buy"
    SELL = "sell"



""" Model for SSI broker """
from typing import Optional
from ..utils import BaseModel


class SSIPlaceOrderRequestModel(BaseModel):
    account: Optional[str] = None
    requestID: Optional[str] = None
    instrumentID: Optional[str] = None
    market: Optional[str] = None
    buySell: Optional[str] = None
    orderType: Optional[str] = None
    price: Optional[float] = 0
    quantity: Optional[int] = 0
    stopOrder: bool = False
    stopPrice: Optional[float] = 0
    stopType: Optional[str] = None
    stopStep: Optional[float] = 0
    lossStep: Optional[float] = 0
    profitStep: Optional[float] = 0
    channelID: Optional[str] = 'TA'
    code: Optional[str] = None
    deviceId: Optional[str] = None
    userAgent: Optional[str] = 'vBroker'


class SSIModifyOrderRequestModel(BaseModel):
    account: Optional[str] = None
    requestID: Optional[str] = None
    orderID: Optional[str] = None
    marketID: Optional[str] = None
    instrumentID: Optional[str] = None
    buySell: Optional[str] = None
    orderType: Optional[str] = None
    price: Optional[float] = 0
    quantity: Optional[int] = 0
    code: Optional[str] = None
    deviceId: Optional[str] = None
    userAgent: Optional[str] = 'vBroker'


class SSICancelOrderRequestModel(BaseModel):
    account: Optional[str] = None
    requestID: Optional[str] = None
    orderID: Optional[str] = None
    marketID: Optional[str] = None
    instrumentID: Optional[str] = None
    buySell: Optional[str] = None
    code: Optional[str] = None
    deviceId: Optional[str] = None
    userAgent: Optional[str] = 'vBroker'



""" Constant for SSI datafeed """
API_URL = "https://fc-tradeapi.ssi.com.vn"
API_PAPER_URL = "https://fc-paperapi.ssi.com.vn"
ENDPOINT_OTP = "api/v2/Trading/GetOTP"
ENDPOINT_AUTH = "api/v2/Trading/AccessToken"

ENDPOINT_EQUITY_NEW_ORDER = "api/v2/Trading/NewOrder"
ENDPOINT_EQUITY_MODIFY_ORDER = "api/v2/Trading/ModifyOrder"
ENDPOINT_EQUITY_CANCEL_ORDER = "api/v2/Trading/CancelOrder"
ENDPOINT_EQUITY_POSITION = "api/v2/Trading/stockPosition"
ENDPOINT_EQUITY_ACCOUNT_BALANCE = "api/v2/Trading/cashAcctBal"

ENDPOINT_DERIVATIVE_NEW_ORDER = "api/v2/Trading/derNewOrder"
ENDPOINT_DERIVATIVE_MODIFY_ORDER = "api/v2/Trading/derModifyOrder"
ENDPOINT_DERIVATIVE_CANCEL_ORDER = "api/v2/Trading/derCancelOrder"
ENDPOINT_DERIVATIVE_POSITION = "api/v2/Trading/derivPosition"
ENDPOINT_DERIVATIVE_ACCOUNT_BALANCE = "api/v2/Trading/derivAcctBal"

ENDPOINT_ORDER_HISTORY = "api/v2/Trading/orderHistory"
ENDPOINT_ORDERBOOK = "api/v2/Trading/orderBook"
ENDPOINT_MAX_BUY_QUANTITY = "api/v2/Trading/maxBuyQty"
ENDPOINT_MAX_SELL_QUANTITY = "api/v2/Trading/maxSellQty"

HUB_URL = "wss://fc-tradehub.ssi.com.vn//v2.0/signalr"
HUB_PAPER_URL = "wss://fc-paperhub.ssi.com.vn//v2.0/signalr"
HUB = "BroadcastHubV2"



from .datetime_handler import split_date_range, convert_timestamp_to_datetime  # noqa: F401
from .enum_handler import EnumHandler  # noqa: F401
from .model_handler import BaseModel, AliasChoices, Field  # noqa: F401
from .request_handler import request_handler  # noqa: F401
from .socket_handler import SocketListener  # noqa: F401
from .jwt_handler import jwt_handler  # noqa: F401
from .sign_data_handler import sign  # noqa: F401



import json
import os.path
from datetime import datetime
from nanoid import generate

from .constant import (
    API_URL,
    ENDPOINT_OTP,
    ENDPOINT_AUTH,

    ENDPOINT_EQUITY_NEW_ORDER,
    ENDPOINT_EQUITY_MODIFY_ORDER,
    ENDPOINT_EQUITY_CANCEL_ORDER,
    ENDPOINT_EQUITY_POSITION,
    ENDPOINT_EQUITY_ACCOUNT_BALANCE,

    ENDPOINT_DERIVATIVE_NEW_ORDER,
    ENDPOINT_DERIVATIVE_MODIFY_ORDER,
    ENDPOINT_DERIVATIVE_CANCEL_ORDER,
    ENDPOINT_DERIVATIVE_POSITION,
    ENDPOINT_DERIVATIVE_ACCOUNT_BALANCE,

    ENDPOINT_ORDER_HISTORY,
    ENDPOINT_ORDERBOOK,
    ENDPOINT_MAX_BUY_QUANTITY,
    ENDPOINT_MAX_SELL_QUANTITY
)
from .model import (
    SSIPlaceOrderRequestModel,
    SSIModifyOrderRequestModel,
    SSICancelOrderRequestModel
)
from .enum import SSISideEnum, SSIOrderStatusEnum

from ..model import vBrokerOrder
from ..interface_broker_api import IBrokerAPI
from ..utils import (
    jwt_handler,
    request_handler,
    sign,
    split_date_range,
    convert_timestamp_to_datetime
)
from ..enum_broker import SideEnum, OrderStatusEnum


class SSIBrokerAPI(IBrokerAPI):
    def __init__(self, config):
        super().__init__(config)
        self.url_auth: str = "/".join([API_URL, ENDPOINT_AUTH])
        self.url_otp: str = "/".join([API_URL, ENDPOINT_OTP])
        self.url_equity_new_order: str = "/".join([API_URL, ENDPOINT_EQUITY_NEW_ORDER])
        self.url_equity_modify_order: str = "/".join([API_URL, ENDPOINT_EQUITY_MODIFY_ORDER])
        self.url_equity_cancel_order: str = "/".join([API_URL, ENDPOINT_EQUITY_CANCEL_ORDER])
        self.url_equity_position: str = "/".join([API_URL, ENDPOINT_EQUITY_POSITION])
        self.url_equity_account_balance: str = "/".join([API_URL, ENDPOINT_EQUITY_ACCOUNT_BALANCE])
        self.url_derivative_new_order: str = "/".join([API_URL, ENDPOINT_DERIVATIVE_NEW_ORDER])
        self.url_derivative_modify_order: str = "/".join(
            [API_URL, ENDPOINT_DERIVATIVE_MODIFY_ORDER]
        )
        self.url_derivative_cancel_order: str = "/".join(
            [API_URL, ENDPOINT_DERIVATIVE_CANCEL_ORDER]
        )
        self.url_derivative_position: str = "/".join([API_URL, ENDPOINT_DERIVATIVE_POSITION])
        self.url_derivative_account_balance: str = "/".join(
            [API_URL, ENDPOINT_DERIVATIVE_ACCOUNT_BALANCE]
        )
        self.url_order_history: str = "/".join([API_URL, ENDPOINT_ORDER_HISTORY])
        self.url_orderbook: str = "/".join([API_URL, ENDPOINT_ORDERBOOK])
        self.url_max_buy_quantity: str = "/".join([API_URL, ENDPOINT_MAX_BUY_QUANTITY])
        self.url_max_sell_quantity: str = "/".join([API_URL, ENDPOINT_MAX_SELL_QUANTITY])
        self.__equity_market_id: str = "VN"
        self.__derivative_market_id: str = "VNFE"
        self.__session_file: str = f"vbroker-ssi-{self.config.ssi_broker_account}.session"
        self.__number: str = "0123456789"
        self.__device_id: str = ":".join([str(i) for i in range(11, 17)])
        self.__headers: dict = {
            "Content-Type": "application/json",
            "Accept": "application/json"
        }
        self.__token: str = None
        self.wait: int = 1  # wait 1 second before sending request

    def get_otp(self) -> str:
        """
        Retrieves the OTP for authentication.
        Returns:
            str: The OTP.
        """
        data: dict = {}
        data.update(
            consumerID=self.config.ssi_broker_id,
            consumerSecret=self.config.ssi_broker_secret
        )
        res = request_handler.post(
            url=self.url_otp, headers=self.__headers, data=data, limit=self.wait
        )
        return f"GET {self.url_otp}: {res}"

    def get_token(self) -> str:
        """
        Retrieves the access token for authentication.
        Returns:
            str: The access token.
        """
        if os.path.exists(self.__session_file):
            with open(self.__session_file, "r") as file:
                self.__token = file.read().strip()

        if jwt_handler.is_expired(bearer_token=self.__token):
            data: dict = {}
            data.update(
                consumerID=self.config.ssi_broker_id,
                consumerSecret=self.config.ssi_broker_secret,
                twoFactorType=1,
                code=self.otp,
                isSave=True
            )
            res = request_handler.post(
                url=self.url_auth, headers=self.__headers, data=data, limit=self.wait
            )
            if res.get("status") == 200:
                access_token = " ".join(["Bearer", res.get("data").get("accessToken")])
                print(access_token)
                with open(self.__session_file, "w") as file:
                    file.write(access_token)
            else:
                access_token = None
                print(f"[vBroker] Failed to get access token.: {res}")
            self.__token = access_token
        return self.__token

    def __get_orderbook_history(
        self, account_no: str, from_date: str, to_date: str, wait: int = 1
    ) -> dict:
        """
        Retrieves the order history for a given account within a specified date range.
        Args:
            account_no (str): The account number for which to retrieve the order history.
            from_date (str): The start date of the date range in the format "YYYY-MM-DD".
            to_date (str): The end date of the date range in the format "YYYY-MM-DD".
        Returns:
            dict: The order history data as a dictionary.
        """
        if from_date:
            from_date = datetime.strptime(from_date, "%Y-%m-%d").strftime("%d/%m/%Y")
        if to_date:
            to_date = datetime.strptime(to_date, "%Y-%m-%d").strftime("%d/%m/%Y")
        self.__headers.update({"Authorization": self.get_token()})
        data: dict = {}
        data.update(
            account=account_no,
            startDate=from_date,
            endDate=to_date
        )
        res = request_handler.get(
            url=self.url_order_history, headers=self.__headers, params=data, limit=wait
        )
        if res.get("status") == 200:
            return res.get("data").get("orderHistories")
        else:
            print(f"[vBroker] {res}")
            return []

    def __get_orderbook(self, account_no: str) -> dict:
        """
        Retrieves the orderbook for a specific account.
        Args:
            account_no (str): The account number.
        Returns:
            dict: The orderbook data.
        Raises:
            None
        """
        self.__headers.update({"Authorization": self.get_token()})
        data: dict = {}
        data.update(
            account=account_no
        )
        res = request_handler.get(
            url=self.url_orderbook, headers=self.__headers, params=data, limit=self.wait
        )
        if res.get("status") == 200:
            return res.get("data").get("order")
        else:
            print(f"[vBroker] {res}")
            return []

    def get_max_buy_quantity(self, account_no: str, instrument: str, price: float) -> dict:
        """
        Retrieves the maximum buy quantity for a specific account.
        Args:
            account_no (str): The account number.
        Returns:
            dict: The maximum buy quantity data.
        Raises:
            None
        """
        self.__headers.update({"Authorization": self.get_token()})
        data: dict = {}
        data.update(
            account=account_no,
            instrumentID=instrument,
            price=price
        )
        res = request_handler.get(
            url=self.url_max_buy_quantity, headers=self.__headers, params=data, limit=self.wait
        )
        return res

    def get_max_sell_quantity(self, account_no: str, instrument: str, price: float) -> dict:
        """
        Retrieves the maximum sell quantity for a specific account.
        Args:
            account_no (str): The account number.
        Returns:
            dict: The maximum sell quantity data.
        Raises:
            None
        """
        self.__headers.update({"Authorization": self.get_token()})
        data: dict = {}
        data.update(
            account=account_no,
            instrumentID=instrument,
            price=price
        )
        res = request_handler.get(
            url=self.url_max_sell_quantity, headers=self.__headers, params=data, limit=self.wait
        )
        return res

    # EQUITY
    def __get_equity_positions(self, account_no: str) -> dict:
        """
        Retrieves the equity positions for a given account.
        Args:
            account_no (str): The account number for which to retrieve equity positions.
        Returns:
            dict: A dictionary containing the equity positions.
        Raises:
            None
        """
        self.__headers.update({"Authorization": self.get_token()})
        data: dict = {}
        data.update(
            account=account_no
        )
        res = request_handler.get(
            url=self.url_equity_position, headers=self.__headers, params=data, limit=self.wait
        )
        return res

    def __get_equity_account_balance(self, account_no: str) -> dict:
        """
        Retrieves the equity account balance for a given account.
        Args:
            account_no (str): The account number for which to retrieve the equity account balance.
        Returns:
            dict: A dictionary containing the equity account balance.
        Raises:
            None
        """
        self.__headers.update({"Authorization": self.get_token()})
        data: dict = {}
        data.update(
            account=account_no
        )
        res = request_handler.get(
            url=self.url_equity_account_balance, headers=self.__headers,
            params=data, limit=self.wait
        )
        return res

    def __place_equity_order(
        self,
        account_no: str, side: str, instrument: str, quantity: int, price: float, order_type: str
    ) -> dict:
        data = SSIPlaceOrderRequestModel(
            account=account_no,
            requestID=generate(alphabet=self.__number, size=8),
            instrumentID=instrument,
            market=self.__equity_market_id,
            buySell=side,
            orderType=order_type,
            price=price,
            quantity=quantity,
            deviceId=self.__device_id
        ).model_dump()
        data_signed = sign(json.dumps(data), self.config.ssi_broker_private_key)
        self.__headers.update({
            "Authorization": self.get_token(),
            "X-Signature": data_signed,
        })
        res = request_handler.post(
            url=self.url_equity_new_order, headers=self.__headers, data=data, limit=self.wait
        )
        return res

    def __modify_equity_order(
        self, account_no: str, order_id: str,
        side: str, instrument: str, quantity: int, price: float, order_type: str
    ) -> dict:
        data = SSIModifyOrderRequestModel(
            account=account_no,
            requestID=generate(alphabet=self.__number, size=8),
            orderID=order_id,
            marketID=self.__equity_market_id,
            instrumentID=instrument,
            buySell=side,
            orderType=order_type,
            price=price,
            quantity=quantity,
            deviceId=self.__device_id
        ).model_dump()
        data_signed = sign(json.dumps(data), self.config.ssi_broker_private_key)
        self.__headers.update({
            "Authorization": self.get_token(),
            "X-Signature": data_signed,
        })
        res = request_handler.post(
            url=self.url_equity_modify_order, headers=self.__headers, data=data, limit=self.wait
        )
        return res

    def __cancel_equity_order(
        self, account_no: str, order_id: str, instrument: str, side: str
    ) -> dict:
        data = SSICancelOrderRequestModel(
            account=account_no,
            requestID=generate(alphabet=self.__number, size=8),
            orderID=order_id,
            marketID=self.__equity_market_id,
            instrumentID=instrument,
            buySell=side,
            deviceId=self.__device_id
        ).model_dump()
        data_signed = sign(json.dumps(data), self.config.ssi_broker_private_key)
        self.__headers.update({
            "Authorization": self.get_token(),
            "X-Signature": data_signed,
        })
        res = request_handler.post(
            url=self.url_equity_cancel_order, headers=self.__headers, data=data, limit=self.wait
        )
        return res

    # DERIVATIVES
    def __get_derivative_positions(self, account_no: str) -> dict:
        """
        Retrieves the derivative positions for a given account.
        Args:
            account_no (str): The account number for which to retrieve derivative positions.
        Returns:
            dict: A dictionary containing the derivative positions.
        Raises:
            None
        """
        self.__headers.update({"Authorization": self.get_token()})
        data: dict = {}
        data.update(
            account=account_no,
            querySummary=True
        )
        res = request_handler.get(
            url=self.url_derivative_position, headers=self.__headers, params=data, limit=self.wait
        )
        return res

    def __get_derivative_account_balance(self, account_no: str) -> dict:
        """
        Retrieves the derivative account balance for a given account.
        Args:
            account_no (str): The account number for which to retrieve
                            the derivative account balance.
        Returns:
            dict: A dictionary containing the derivative account balance.
        Raises:
            None
        """
        self.__headers.update({"Authorization": self.get_token()})
        data: dict = {}
        data.update(
            account=account_no
        )
        res = request_handler.get(
            url=self.url_derivative_account_balance, headers=self.__headers,
            params=data, limit=self.wait
        )
        return res

    def __place_derivative_order(
        self,
        account_no: str, side: str, instrument: str, quantity: int, price: float, order_type: str
    ) -> dict:
        data = SSIPlaceOrderRequestModel(
            account=account_no,
            requestID=generate(alphabet=self.__number, size=8),
            instrumentID=instrument,
            market=self.__derivative_market_id,
            buySell=side,
            orderType=order_type,
            price=price,
            quantity=quantity,
            deviceId=self.__device_id
        ).model_dump()
        data_signed = sign(json.dumps(data), self.config.ssi_broker_private_key)
        self.__headers.update({
            "Authorization": self.get_token(),
            "X-Signature": data_signed,
        })
        res = request_handler.post(
            url=self.url_derivative_new_order, headers=self.__headers, data=data, limit=self.wait
        )
        return res

    def __modify_derivative_order(
        self, account_no: str, order_id: str,
        side: str, instrument: str, quantity: int, price: float, order_type: str
    ) -> dict:
        data = SSIModifyOrderRequestModel(
            account=account_no,
            requestID=generate(alphabet=self.__number, size=8),
            orderID=order_id,
            marketID=self.__derivative_market_id,
            instrumentID=instrument,
            buySell=side,
            orderType=order_type,
            price=price,
            quantity=quantity,
            deviceId=self.__device_id
        ).model_dump()
        data_signed = sign(json.dumps(data), self.config.ssi_broker_private_key)
        self.__headers.update({
            "Authorization": self.get_token(),
            "X-Signature": data_signed,
        })
        res = request_handler.post(
            url=self.url_derivative_modify_order, headers=self.__headers, data=data, limit=self.wait
        )
        return res

    def __cancel_derivative_order(
        self, account_no: str, order_id: str, instrument: str, side: str
    ) -> dict:
        data = SSICancelOrderRequestModel(
            account=account_no,
            requestID=generate(alphabet=self.__number, size=8),
            orderID=order_id,
            marketID=self.__derivative_market_id,
            instrumentID=instrument,
            buySell=side,
            deviceId=self.__device_id
        ).model_dump()
        data_signed = sign(json.dumps(data), self.config.ssi_broker_private_key)
        self.__headers.update({
            "Authorization": self.get_token(),
            "X-Signature": data_signed,
        })
        res = request_handler.post(
            url=self.url_derivative_cancel_order, headers=self.__headers, data=data, limit=self.wait
        )
        return res

    def get_ordebbook(self, account_no: str, from_date: str = None, to_date: str = None) -> dict:
        if all([
            from_date is None,
            to_date is None
        ]):
            from_date = to_date = datetime.now().strftime("%Y-%m-%d")
        _orderbook: list = []
        _split_date_range = split_date_range(from_date, to_date) or [(from_date, to_date)]
        for _f_date, _t_date in _split_date_range:
            _order = self.__get_orderbook_history(account_no, _f_date, _t_date, wait=10)
            if _order:
                _orderbook += _order
        return [
            vBrokerOrder(
                **i,
                account_no=account_no,
                side=SideEnum[SSISideEnum[i.get("buySell")].value],
                status=OrderStatusEnum[SSIOrderStatusEnum[i.get("orderStatus")].value],
                os_quantity=i.get("quantity") - i.get("filledQty"),
                input_time=convert_timestamp_to_datetime(
                    int(i.get("inputTime"))/1000
                ),
                modified_time=convert_timestamp_to_datetime(
                    int(i.get("modifiedTime"))/1000
                ),
            ) for i in _orderbook
        ]

    def get_positions(self, account_no: str, is_equity: bool = True) -> dict:
        if is_equity:
            return self.__get_equity_positions(account_no)
        else:
            return self.__get_derivative_positions(account_no)

    def get_balance(self, account_no: str, is_equity: bool = True) -> dict:
        if is_equity:
            return self.__get_equity_account_balance(account_no)
        else:
            return self.__get_derivative_account_balance(account_no)

    def place_order(
        self, account_no: str, side: str, instrument: str,
        quantity: int, price: float | str, is_equity: bool = True
    ) -> dict:
        if isinstance(price, str):
            order_type = price
            price = 0
        else:
            order_type = "LO"
        side = SSISideEnum[SideEnum[side.upper()].value].value
        if is_equity:
            res = self.__place_equity_order(
                account_no=account_no,
                side=side,
                instrument=instrument,
                quantity=quantity,
                price=price,
                order_type=order_type
            )
        else:
            res = self.__place_derivative_order(
                account_no=account_no,
                side=side,
                instrument=instrument,
                quantity=quantity,
                price=price,
                order_type=order_type
            )
        order_info = res.get("data")
        _unique_id = order_info.get("requestID")
        return vBrokerOrder(
            order_id=None,
            account_no=account_no,
            side=SideEnum[SSISideEnum[side].value],
            instrument=instrument,
            quantity=quantity,
            filled_quantity=0,
            os_quantity=quantity,
            cancelled_quantity=0,
            price=price,
            avg_price=0,
            order_type=order_type,
            unique_id=_unique_id,
            input_time=datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            modified_time=datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            status=OrderStatusEnum.PREPARE,
            message=None
        )

    def modify_order(
        self, account_no: str, order_id: str, side: str, instrument: str,
        quantity: int, price: float | str, is_equity: bool = True
    ) -> dict:
        if isinstance(price, str):
            order_type = price
            price = 0
        else:
            order_type = "LO"
        side = SSISideEnum[SideEnum[side.upper()].value].value
        if is_equity:
            res = self.__modify_equity_order(
                account_no=account_no,
                order_id=order_id,
                side=side,
                instrument=instrument,
                quantity=quantity,
                price=price,
                order_type=order_type
            )
        else:
            res = self.__modify_derivative_order(
                account_no=account_no,
                order_id=order_id,
                side=side,
                instrument=instrument,
                quantity=quantity,
                price=price,
                order_type=order_type
            )
        order_info = res.get("data").get("requestData")
        _unique_id = res.get("data").get("requestID")
        return vBrokerOrder(
            order_id=order_info.get("orderID"),
            account_no=account_no,
            side=SideEnum[SSISideEnum[side].value],
            instrument=instrument,
            quantity=quantity,
            filled_quantity=0,
            os_quantity=quantity,
            cancelled_quantity=0,
            price=price,
            avg_price=0,
            order_type=order_type,
            unique_id=_unique_id,
            input_time=datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            modified_time=datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            status=OrderStatusEnum.WAITING,
            message=None
        )

    def cancel_order(
        self, account_no: str, order_id: str, instrument: str, side: str, is_equity: bool = True
    ) -> dict:
        side = SSISideEnum[SideEnum[side.upper()].value].value
        if is_equity:
            return self.__cancel_equity_order(
                account_no=account_no,
                order_id=order_id,
                instrument=instrument,
                side=side
            )
        else:
            return self.__cancel_derivative_order(
                account_no=account_no,
                order_id=order_id,
                instrument=instrument,
                side=side
            )



""" Model for vBroker """
from typing import Optional
from .utils import BaseModel, Field, AliasChoices


class vBrokerOrder(BaseModel):
    account_no: Optional[str] = Field(validation_alias=AliasChoices('account_no', 'account'))
    order_id: Optional[str] = Field(validation_alias=AliasChoices('order_id', 'orderID'))
    unique_id: Optional[str] = Field(
        validation_alias=AliasChoices('unique_id', 'uniqueID', 'requestsID')
    )
    instrument: Optional[str] = Field(validation_alias=AliasChoices('instrument', 'instrumentID'))
    side: Optional[str] = Field(validation_alias=AliasChoices('side', 'buySell'))
    order_type: Optional[str] = Field(validation_alias=AliasChoices('order_type', 'orderType'))
    price: Optional[float] = Field(validation_alias=AliasChoices('price', 'Price'))
    avg_price: Optional[float] = Field(validation_alias=AliasChoices('avg_price', 'avgPrice'))
    quantity: Optional[int] = Field(validation_alias=AliasChoices('quantity', 'Quantity'))
    filled_quantity: Optional[int] = Field(
        validation_alias=AliasChoices('filled_quantity', 'filledQty')
    )
    os_quantity: Optional[int] = Field(
        validation_alias=AliasChoices('os_quantity', 'osQty')
    )
    cancelled_quantity: Optional[int] = Field(
        validation_alias=AliasChoices('cancelled_quantity', 'cancelQty')
    )
    status: Optional[str] = Field(validation_alias=AliasChoices('status', 'orderStatus'))
    input_time: Optional[str] = Field(validation_alias=AliasChoices('input_time', 'inputTime'))
    modified_time: Optional[str] = Field(
        validation_alias=AliasChoices('modified_time', 'modifiedTime')
    )
    message: Optional[str] = Field(validation_alias=AliasChoices('message', 'rejectReason'))



import time
import requests


class RequestHandler:
    def __init__(self) -> None:
        self.__timeout: int = 10

    def get(self, url: str, headers: dict, params: dict, limit: int = 0) -> dict:
        try:
            print(f"[vBroker] Call ~> GET: {url}: {params}")
            res = requests.get(url, headers=headers, params=params, timeout=self.__timeout)
            if limit:
                time.sleep(limit)
            res.raise_for_status()
            return res.json()
        except requests.RequestException as e:
            raise e

    def post(self, url: str, headers: dict, data: dict = {}, limit: int = 0) -> dict:
        try:
            print(f"[vBroker] Call ~> POST: {url} data: {data}")
            if data:
                res = requests.post(url, headers=headers, json=data, timeout=self.__timeout)
            else:
                res = requests.post(url, headers=headers, timeout=self.__timeout)
            if limit:
                time.sleep(limit)
            res.raise_for_status()
            return res.json()
        except requests.RequestException as e:
            raise e


request_handler = RequestHandler()



""" This module defines the interface for a broker HUB. """
from abc import ABC, abstractmethod


class IBrokerHUB(ABC):
    def __init__(self, api):
        self.api = api
        self.__token: str = None

    @property
    def token(self) -> str:
        """
        Returns the token used for authentication.
        Returns:
            str: The authentication token.
        """
        return self.__token

    @token.setter
    def token(self, value: str):
        self.__token = value

    @abstractmethod
    async def listen(self, args, on_message):
        """
        Listens for incoming messages from the broker.
        Args:
            args: The arguments for the listen method.
            on_message: The callback function to be executed when a message is received.
        Returns:
            NotImplemented
        """
        return NotImplemented



import json
import base64
from Crypto.Hash import SHA256
from Crypto.PublicKey import RSA
from Crypto.Signature import pkcs1_15
from xmljson import badgerfish as bf
from xml.etree.ElementTree import fromstring


def get_rsa_key(private_key: str):
    key = base64.b64decode(private_key.encode('utf-8'))
    key_json = json.loads(json.dumps(bf.data(fromstring(key))))
    cc = key_json['RSAKeyValue']
    n = int.from_bytes(base64.b64decode(cc['Modulus']['$']), byteorder='big')
    e = int.from_bytes(base64.b64decode(cc['Exponent']['$']), byteorder='big')
    d = int.from_bytes(base64.b64decode(cc['D']['$']), byteorder='big')
    p = int.from_bytes(base64.b64decode(cc['P']['$']), byteorder='big')
    q = int.from_bytes(base64.b64decode(cc['Q']['$']), byteorder='big')
    return RSA.construct((n, e, d, p, q))


def sign(data: str, private_key: str):
    rsa_key = get_rsa_key(private_key)
    h = SHA256.new(data.encode('utf-8'))
    return pkcs1_15.new(rsa_key).sign(h).hex()



from pydantic import BaseModel as BM, ConfigDict, Field, AliasChoices  # noqa: F401


class BaseModel(BM):
    model_config: ConfigDict = ConfigDict(
        use_enum_values=True,
        validate_return=True
    )
